#include "frameBuffer.h"

#include "FL/gl.h"
#include <iostream.h>

#include "globals.h"
#include <math.h>
#include <float.h>

#include <tiffio.h>
#include <fstream.h>


FrameBuffer::FrameBuffer(int u0, int v0, int _w, int _h) : 
  w(_w), h(_h), Fl_Gl_Window(u0, v0, _w, _h, 0) {

  pix = new unsigned int[w*h];
  zb = new float[w*h];

}

void FrameBuffer::draw() {

  if (pix)
    glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);
  else
    cerr << "attempted to post bad pixels in FrameBuffer::PostSWFB()" << endl;

}

void FrameBuffer::show() {

  if (pix)
    size(w, h);
  Fl_Gl_Window::show();
  
}

int FrameBuffer::handle(int event) {

  switch (event) {
  case FL_RELEASE: {
    int mouseu = Fl::event_x();
    int mousev = Fl::event_y();
    cerr << "INFO: released mouse button at: " << 
      mouseu << " " << mousev << endl;
    break;
  }
  case FL_MOVE: {
    int mouseu = Fl::event_x();
    int mousev = Fl::event_y();
    scene->MouseMove(mouseu, mousev);
    break;
  }
  case FL_KEYBOARD: {
    int key = Fl::event_key();
    scene->KeyPress(key);
    break;
  }
  default: {
    // events I don't handle
  }
  }

  return 1;

}

void FrameBuffer::Set(int u, int v, unsigned int color) {

  pix[(h-1-v)*w+u] = color;

}

unsigned int FrameBuffer::Get(int u, int v) {

  return pix[(h-1-v)*w+u];

}

void FrameBuffer::SetGuarded(int u, int v, unsigned int color) {

  if (u > 0 && u < w && v > 0 && v < h) {
    Set(u, v, color);
  }

}

void FrameBuffer::Clear(unsigned int clearColor) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      Set(u, v, clearColor);
      zb[(h-1-v)*w+u] = 0.0f;
    }
  }

}

void FrameBuffer::MakeChecker(int checkerWidth, unsigned int *colors) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      unsigned int currColor = colors[0];
      int cu = u/checkerWidth;
      int cv = v/checkerWidth;
      if ((cu+cv)%2) {
        currColor = colors[1];
      }
      Set(u, v, currColor);
    }
  }


}

void FrameBuffer::Setf(float uf, float vf, unsigned color) {

  int u = (int)uf;
  int v = (int)vf;
  SetGuarded(u, v, color);

}

void FrameBuffer::DrawPointDiagonalCross(float uf, float vf, 
  int size, unsigned int color) {


  int u = (int)uf;
  int v = (int)vf;
  for (int i = 0; i < size; i++) {
    SetGuarded(u-size/2+i, v-size/2+i, color);
    SetGuarded(u+size/2-i, v-size/2+i, color);
  }

}

void FrameBuffer::DrawPointSquare(float uf, float vf, 
  int size, unsigned int color) {


  int u = (int)uf;
  int v = (int)vf;
  for (int cv = v-size/2; cv <= v+size/2; cv++) {
    for (int cu = u-size/2; cu <= u+size/2; cu++) {
      SetGuarded(cu, cv, color);
    }
  }
}

void FrameBuffer::DrawSegment(float u0f, float v0f, float u1f, float v1f,
  unsigned int color) {

  float dv = v1f-v0f;
  float du = u1f-u0f;
  int nSteps;
  if (fabsf(dv) > fabsf(du)) {
    nSteps = (int) ceilf(fabsf(dv));
  }
  else {
    nSteps = (int) ceilf(fabsf(du));
  }

  float curruf = u0f;
  float currvf = v0f;
  for (int i = 0; i <= nSteps; i++) {
    Setf(curruf, currvf, color);
    curruf += du/(float)nSteps;
    currvf += dv/(float)nSteps;
  }

}

void FrameBuffer::DrawPointDiagonalCross3D(V3& point, PHC &phc, 
  int size, unsigned int color) {

  V3 projP;
  if (!phc.Project(point, projP))
    return;

  DrawPointDiagonalCross(projP[0], projP[1], size, color);

}

void FrameBuffer::DrawPointSquare3D(V3& point, PHC &phc, 
  int size, unsigned int color) {

  V3 projP;
  if (!phc.Project(point, projP))
    return;

  DrawPointSquare(projP[0], projP[1], size, color);

}


void FrameBuffer::DrawSegment3D(V3 point0, V3 point1, PHC &phc,
  unsigned int color) {


  V3 projP[2];
  if (!phc.Project(point0, projP[0]))
    return;
  if (!phc.Project(point1, projP[1]))
    return;

  DrawSegment(projP[0][0], projP[0][1], projP[1][0], projP[1][1],
    color);

}

float& FrameBuffer::ZB(int u, int v) {

  return zb[(h-1-v)*w+u];

}

void FrameBuffer::Rotate(PHC& phc0, PHC& phc1, FrameBuffer& I1) {

  M33 M0 = phc0.GetCamMat();
  M33 M1 = phc1.GetCamMat();
  M33 R = M0.Inverted()*M1;
  for (int v1 = 0; v1 < I1.h; v1++) {
    for (int u1 = 0; u1 < I1.w; u1++) {
      V3 pixv(0.5f+(float)u1, 0.5f+(float)v1, 1.0f);
      float w0 = R[2]*pixv;
      float u0 = (R[0]*pixv) / w0;
      float v0 = (R[1]*pixv) / w0;
      if (u0 < 0.0f || u0 > (float)w)
        continue;
      if (v0 < 0.0f || v0 > (float)h)
        continue;
      I1.Set(u1, v1, Get((int)u0, (int)v0));
    }
  }

}


void FrameBuffer::RotateFM(PHC& phc0, PHC& phc1, FrameBuffer& I1) {

  M33 M0 = phc0.GetCamMat();
  M33 M1 = phc1.GetCamMat();
  M33 R = M1.Inverted()*M0;
  for (int v0 = 0; v0 < h; v0++) {
    for (int u0 = 0; u0 < w; u0++) {
      V3 pixv(0.5f+(float)u0, 0.5f+(float)v0, 1.0f);
      float w1 = R[2]*pixv;
      float u1 = (R[0]*pixv) / w1;
      float v1 = (R[1]*pixv) / w1;
      if (u1 < 0.0f || u1 > (float)I1.w)
        continue;
      if (v1 < 0.0f || v1 > (float)I1.h)
        continue;
      I1.Set((int)u1, (int)v1, Get(u0, v0));
    }
  }
}


void FrameBuffer::Set(FrameBuffer *src, int u0, int v0) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      Set(u, v, src->Get(u0+u, v0+v));
    }
  }

}


int FrameBuffer::Load(char *tiffFileName) {
  
  if (!strcmp(tiffFileName + strlen(tiffFileName)-4, ".tif")) {
    
    ifstream tst(tiffFileName, ios::nocreate);
    if (tst.fail()) {
      cerr << "INFO: cannot open: " << tiffFileName << endl;
      return 0;
    }
    tst.close();
    
    TIFF* tif = TIFFOpen(tiffFileName, "r");
    
    if (tif) {
      if (pix)
        delete pix;
      
      cerr << "INFO: started loading: " << tiffFileName << endl;
      
      TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
      TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
      pix = new unsigned int[w*h];
      size(w, h);
      
      if (!TIFFReadRGBAImage(tif, w, h, (unsigned long*) pix, 0)) {
        delete pix;
        pix = 0;
        w = h = 0;
        return 0;
      }
      
      TIFFClose(tif);
      label(tiffFileName);
      cerr << "INFO: done loading: " << tiffFileName << endl;
      return 1;
    }
    return 0;
  }
  cerr << "INFO: do not know how to open *" << tiffFileName + strlen(tiffFileName)-4 << endl;
  return 0;
}

